
import time
from enum import Enum

from termcolor import colored
import datetime

DUMP_TO_FILE = False

class LogSeverity(Enum):
    SUCCESS = "Success"
    WARNING = "Warning"
    INFO = "Info"
    ERROR = "Error"
    FATAL = "Fatal"

def getColor(sev):
    return {
        'Success' : 'green',
        'Warning': 'yellow',
        'Info': 'blue',
        'Error': 'magenta',
        'Fatal': 'red'
    }.get(sev, 'white')


def log(message, log_sev=LogSeverity.INFO, time=time.time()):
    color = getColor(log_sev.value)
    adapted_time = datetime.datetime.fromtimestamp(time)
    str_to_print = colored("[%s] %s: %s" % (log_sev.value, adapted_time, message), color)
    print(str_to_print)
    if DUMP_TO_FILE:
        filename = "%d-%d-%d.log" % (adapted_time.year,adapted_time.month,adapted_time.day)
        f = open(filename, "a+")
        f.write(str_to_print+"\n")
        f.close()


from .service import Service



from ..utils.Logger import log
from ..utils.Utils import execute_shell_command


class BatteryStatsService(Service):
	"""docstring for BatteryStatsService"""
	def __init__(self,boot_time=0, output_res_folder="batterystats"):
		Service.__init__(self,output_res_folder)
		self.boot_time = boot_time

	def config(self,**kwargs):
		pass

	def init(self,boot_time=0,**kwargs):
		self.boot_time = boot_time
		self.clean()

	def start(self):
		execute_shell_command("adb shell dumpsys batterystats --reset")

	def stop(self,fil=None):
		
		if fil is None:
			uid_file = execute_shell_command("date +%s")[1].strip()
		filename = self.results_dir + "/bstats-%s-%s.log" % (uid_file, str(self.boot_time))
		execute_shell_command("adb shell dumpsys batterystats --history > %s" % filename)
		#print(colored("Output file  %s" % filename, "green"))
		log("Output file  %s" % filename)
		return filename

	def clean(self):
		execute_shell_command("find %s -type f  | xargs rm " % self.results_dir)







from __future__ import absolute_import

from .service import Service
import time
import os

from manafa.utils.Utils import execute_shell_command, get_resources_dir

RESOURCES_DIR = get_resources_dir()

class PerfettoService(Service):
	"""docstring for BatteryStatsService"""
	def __init__(self, boot_time=0, output_res_folder="perfetto"):
		Service.__init__(self, output_res_folder)
		self.boot_time = boot_time

	def config(self, **kwargs):
		pass

	def init(self,boot_time=0,**kwargs):
		self.boot_time = boot_time
		self.clean()

	def start(self):
		execute_shell_command("adb shell perfetto -o /data/misc/perfetto-traces/trace freq  -t 1h --background ")
	
	def stop(self, file_id=None):
		if file_id is None:
			file_id = execute_shell_command("date +%s")[1].strip()
		#executeShCommand("adb shell su -c \"killall perfetto\"")
		execute_shell_command("adb shell killall perfetto")
		time.sleep(1)
		execute_shell_command("adb pull /data/misc/perfetto-traces/trace " + ("%strace" % self.results_dir ) + "-" + file_id + "-" + str(self.boot_time))
		return self.export()

	def export(self):
		last_exported = ""
		for f in os.listdir(self.results_dir):
			res, o, e = execute_shell_command("chmod +x %s/traceconv ; %s/traceconv systrace %s/%s %s/%s.systrace" %(RESOURCES_DIR,RESOURCES_DIR,self.results_dir,f,self.results_dir,f) )
			last_exported = "%s/%s.systrace" %(self.results_dir,f)
		return last_exported

	def clean(self):
		execute_shell_command("find %s -type f  | xargs rm " % self.results_dir)





import re
from manafa.powerProfile.PowerProfile import PowerProfile

x="""import time
import subprocess
def executeShCommand(command):

    pipes = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    std_out, std_err = pipes.communicate()
    output = std_out.decode("utf-8").lower()
    if pipes.returncode != 0:
        err_msg = "%s. Code: %s" % (std_err.decode('utf-8').strip(), pipes.returncode)
        print("error executing command %s" % command)
        print(err_msg)
        return -1    
    elif len(std_err)==0:
        return output


def epochToDate(ts):
	return time.ctime(ts)
###"""

def interpolate(x1: float, x2: float, y1: float, y2: float, x: float):
	"""Perform linear interpolation for x between (x1,y1) and (x2,y2) """
	return ((y2 - y1) * x + x2 * y1 - x1 * y2) / (x2 - x1)  if (x2-x1)>0 else y1
	#print(val)
	#print("---")
	#return val

class PerfettoEvent(object):
	"""docstring for BatteryEvent"""
	def __init__(self, time=0.0, values=[]):
		self.time=time
		self.vals=[]
		for x in values:
			self.vals.append(x)

	def __str__(self):
		return "time:%f vals =  %s , " % (self.time,str(self.vals))

	def __repr__(self):
		return str(self)

	def initAll(self,default_len=8,val=0):
		for x in range(0,default_len):
			if len(self.vals)> x :
				self.vals[x]=val 
			else: 
				self.vals.append(val)

	def update(self, cpu_id,cpu_freq):
		if len(self.vals)> cpu_id:
			self.vals[cpu_id]=cpu_freq 
		else: 
			for x in range(len(self.vals)-1,cpu_id ):
				self.vals.append( cpu_freq)

	
	def calculateCPUsCurrent(self,state, profile):
		total=0
		if state not in ["idle","suspend"]:
			for core_id, freq in enumerate(self.vals):
				bf,aft = profile.getCPUCoreSpeedPair(core_id,freq)
				lin_inter_val = interpolate( bf[0], aft[0], bf[1], aft[1], freq )
				total += lin_inter_val
			total = total / len(self.vals)
		else:
			total = profile.getCPUStateCurrent(state)

		return total / 1000


class PerfettoCPUfreqParser(object):
	def __init__(self, power_profile=None, start_time=0.0, timezone="EST"):
		self.events = []
		self.start_time = start_time
		self.power_profile = self.loadPowerProfile(power_profile) if power_profile is not None else {}

	def loadPowerProfile(self, xml_profile ):
		return PowerProfile(xml_profile)
	
	def parseFile(self, filename):
		with open(filename, 'r') as filehandle:
			lines=filehandle.read().splitlines()
			self.parseHistory(lines)

	def parseHistory(self, lines):
		for line in lines:
			if line.startswith("#"):
				continue
			z = re.match(r"^\s*([^\s]+)\-(\d+)\s*\(\s*(\d+|\-+)\) \[(\d+)\] (\d+|\.+) ([0-9]*\.[0-9]+|[0-9]+)\: (.*)?$",line)
			if z is not None:
				time = float(z.groups()[5])
				time += self.start_time
				ev_pair = self.parseEvent(z.groups()[6])
				if ev_pair is not None:
					cpu_id = ev_pair[0]
					cpu_freq = ev_pair[1]
					self.addEvent(time,cpu_id,cpu_freq)
			else:
				raise Exception ("Error parsing file")
		
	def addEvent(self,time,cpu_id,cpu_freq):
		if len(self.events)==0:
			z = PerfettoEvent(time)
			z.initAll(default_len=8,val=cpu_freq)
			self.events.append(z)
		else:
			last = self.events[-1]
			z = PerfettoEvent(time, last.vals)
			z.update(cpu_id,cpu_freq)
			self.events.append(z)

	def parseEvent(self, ev_str):
		mat = re.match (r"cpu_frequency: state=(\d+) cpu_id=(\d+)", ev_str)
		if mat:
			cpu_id=int(mat.groups()[1])
			cpu_freq=int(mat.groups()[0])
			return cpu_id,cpu_freq
	

#bootTime = float ( executeShCommand ("adb shell cat /proc/stat | grep btime | awk '{print $2}'").strip() )
#print(bootTime)
#print(epochToDate(bootTime))
#x = PerfettoCPUfreqParser(bootTime)
#x.parseFile("/Users/ruirua/repos/petra_like/results/perfetto/trace-1605638909.systrace")
#print(x.events)




from unittest import TestCase


class TestPerfettoService(TestCase):
    def test_start(self):
        self.assertTrue(True)

    def test_stop(self):
        self.assertTrue(True)

    def test_export(self):
        self.assertTrue(True)

    def test_clean(self):
        self.assertTrue(True)




import os, sys

import xml.etree.ElementTree as ET


class PowerProfile(object):
	"""docstring for PowerProfile"""
	def __init__(self,filename):
		self.filename = filename
		self.components={}
		self.read_power_profile()

	def addComponent(self,component):
		if component.name in self.components:
			self.components[component.name].states =  merge_two_dicts( self.components[component.name].states , (component.states))
		else:
			self.components[component.name] = component
		
	def __str__(self):
		return str(self.components)

	def __repr__(self):
		return str(self)				

	def read_power_profile(self):
		try:
			tree = ET.parse(self.filename)
			root = tree.getroot()
		except Exception as e:
			print("Exception: {0}".format(e))
			return
		
		for child in root:
			if child.tag == "item":
				ll = child.attrib['name'].split(".")
				begin_d = self.components
				for at in ll:
					begin_d[at]={} if not at in begin_d else begin_d[at]
					last_b = begin_d
					begin_d = begin_d[at]
				last_b[at]=float(child.text)
			elif child.tag == "array":
				ll = child.attrib['name'].split(".")
				begin_d = self.components
				for at in ll:
					begin_d[at]={} if at not in begin_d else begin_d[at]
					last_b = begin_d
					begin_d = begin_d[at]
				last_b[at]= list( map( lambda xxz : float(xxz.text), child.getchildren() ) )

	
	def getCPUStateCurrent(self,state):
		return self.components["cpu"][state] if state in self.components else self.components["cpu"]["active"]

	# returns pair with closest_val_before_freq,closest_val_after_freq
	def getCPUCoreSpeedPair(self,core_id,core_freq):
		#measured in KHz
		profile_speeds = self.components["cpu"]["speeds"]  if "speeds" in self.components["cpu"]else self.components["cpu"]["core_speeds"] 
		profile_currents =  self.components["cpu"]["active"]  if ("speeds" in self.components["cpu"] and isinstance( self.components["cpu"]["active"] ,list ) ) else self.components["cpu"]["core_power"] 

		if isinstance(profile_speeds, dict ):
			# select respective cluster of core_id	
			core_r = core_id +1
			cluster_cores = self.components["cpu"]["clusters"]["cores"]
			for i,ncores in enumerate(cluster_cores):
				if core_r > ncores:
					core_r= core_r - ncores
				else:
					profile_speeds=self.components["cpu"]["core_speeds"]["cluster%d" % i]	
					profile_currents= self.components["cpu"]["core_power"]["cluster%d" % i]	
					break
		mini_fr = 0
		freq = 0
		# find adequate freq
		for i,f in enumerate(profile_speeds):
			if f > core_freq:
				freq = i
				break
			mini_fr = i
		fst_pair = (profile_speeds[mini_fr], profile_currents[mini_fr])
		snd_pair = (profile_speeds[freq], profile_currents[freq])
		return fst_pair, snd_pair


if __name__ == '__main__':
	if len(sys.argv)>1:
		power_profile=PowerProfile(sys.argv[1])
		print(power_profile)


import sysconfig
from subprocess import Popen, PIPE
from textops import find

DEV=False


def get_resources_dir(packname="manafa", default_res_dir="resources"):
    if DEV:
        return packname + "/" + default_res_dir
    return sysconfig.get_path("purelib") + "/" + packname + "/" + default_res_dir


def get_pack_dir(packname="manafa"):
    if DEV:
        return packname
    return sysconfig.get_path("purelib") + "/" + packname


def get_results_dir(packname="manafa", default_results_dir="results"):
    if DEV:
        return packname + "/" + default_results_dir
    return sysconfig.get_path("purelib") + "/" + packname + "/" + default_results_dir
    #return packname + "/" + default_results_dir


def execute_shell_command(cmd, args=[]):
    command = cmd + " " + " ".join(args) if len(args) > 0 else cmd
    proc = Popen(command, stdout=PIPE, stderr=PIPE,shell=True)
    out, err = proc.communicate()
    return proc.returncode, out.decode("utf-8") , err.decode('utf-8')

def mega_find(basedir, pattern="*", maxdepth=999, mindepth=0, type_file='n'):
    basedir_len = len(basedir.split("/"))
    res = find(basedir, pattern=pattern, only_files=type_file=='f', only_dirs=type_file=='d' )
    # filter by depth
    return list( filter(lambda x : len(x.split("/")) >= basedir_len + mindepth and len(x.split("/")) <= maxdepth + basedir_len , res ) )




import time
from manafa.services.batteryStatsService import BatteryStatsService
from manafa.services.service import *
from manafa.services.perfettoService import PerfettoService
from manafa.perfetto.perfettoParser import PerfettoCPUfreqParser
from manafa.batteryStats.BatteryStatsParser import BatteryStatsParser
import argparse
from manafa.utils.Logger import log, LogSeverity
from manafa.utils.Utils import execute_shell_command, mega_find, get_resources_dir

MANAFA_RESOURCES_DIR = get_resources_dir()

DEFAULT_PROFILE = MANAFA_RESOURCES_DIR + "/profiles/power_profile.xml"
DEFAULT_TIMEZONE = "GMT"

def get_last_boot_time(bts_file=None):
	res, out, err = execute_shell_command("adb shell cat /proc/stat | grep btime | awk '{print $2}'") #executeShCommand("adb shell cat /proc/stat | grep btime | awk '{print $2}'")
	if res != 0 or len(out) == 0:
		log("no device connected. Assuming Boot time of battery stats file", LogSeverity.ERROR)
		flds = bts_file.split("-") if bts_file is not None else []
		if len(flds)>1:
			log("Boot time: " + flds[2], LogSeverity.WARNING )
			return int(flds[2])
		else:
			log("no device connected. Assuming Boot time 0", LogSeverity.WARNING)
			return 0
			#raise Exception("Invalid boot time ")
		#print("[Warning]: no device connected. Assuming Boot time %d" % boot_time)
	return float(out.strip())

def is_float(string):
	try:
		float(string)
	except ValueError:
		return False
	return True

class EManafa(Service):
	"""docstring for GreenStats"""
	def __init__(self, power_profile=None, timezone=None, resources_dir=MANAFA_RESOURCES_DIR):
		Service.__init__(self)
		self.resources_dir = resources_dir
		self.power_profile = power_profile if power_profile is not None else self.inferPowerProfile()
		self.boot_time = 0
		log("Power profile file: " + self.power_profile , LogSeverity.INFO )
		self.batterystats = BatteryStatsService()
		self.perf_events = None
		self.perfetto = PerfettoService()
		self.timezone = timezone if timezone is not None else self.__inferTimezone()
		self.unplugged=False

	def config(self, **kwargs):
		pass

	def init(self):
		self.boot_time = get_last_boot_time()
		self.batterystats.init(boot_time=self.boot_time)
		self.perfetto.init(boot_time=self.boot_time)
		self.__unplug_if_fully_charged()
		
	def start(self):
		self.batterystats.start()
		self.perfetto.start()

	def stop(self):
		b_out = self.batterystats.stop()
		p_out = self.perfetto.stop()
		if self.unplugged:
			self.__plug_back()
		return b_out, p_out
	
	def clean(self):
		self.batterystats.clean()
		self.perfetto.clean()

	def parseResults(self, bts_file="", pf_file=""):
		if bts_file == "" or pf_file == "":
			log("Empty result files", LogSeverity.FATAL)
			raise Exception()
		self.b_time = get_last_boot_time(bts_file)
		self.bat_events = BatteryStatsParser(self.power_profile, timezone=self.timezone)
		self.bat_events.parseFile(bts_file)
		self.perf_events = PerfettoCPUfreqParser(self.power_profile, self.b_time, timezone=self.timezone)
		self.perf_events.parseFile(pf_file)

	# energy calc related stuff
	def getConsumptionInBetween(self, start_time=0, end_time=9905715380):
		total, per_component = self.calculateNonCpuEnergy(start_time, end_time)
		total_cpu = self.calculateCPUEnergy(start_time,end_time)
		per_component['cpu'] += total_cpu
		return total + total_cpu, per_component

	def calculateNonCpuEnergy(self, start_time, end_time):
		c_beg_bef, c_beg_aft = self.__getClosestPair(self.bat_events.events, start_time)
		total = 0
		per_component_consumption = {}
		last_event = self.bat_events.events[c_beg_bef]
		last_time = start_time
		for i, x in enumerate(self.bat_events.events[c_beg_aft:]):
			if x.time > end_time:
				break
			delta_time = abs(x.time - last_time)
			tot_curr, comps_curr = last_event.getCurrentOfBatStatEvent()
			total += tot_curr * (last_event.getVoltageValue()) * delta_time
			for comp, comp_curr in comps_curr.items():
				if comp not in per_component_consumption:
					per_component_consumption[comp] = 0
				if is_float(comp_curr):
					per_component_consumption[comp] += (comp_curr * last_event.getVoltageValue() * delta_time)
			last_event = x
			last_time = x.time

		delta_time = end_time - last_time 
		if delta_time < 0.0:
			log(time.time(), "Error calculating delta (<0) ", LogSeverity.FATAL)
		tot_curr, comps_curr = last_event.getCurrentOfBatStatEvent()
		total += tot_curr * (last_event.getVoltageValue()) * (delta_time)
		for comp, comp_curr in comps_curr.items():
			if comp not in per_component_consumption:
				per_component_consumption[comp] = 0
			if is_float(comp_curr):
				per_component_consumption[comp] += (comp_curr * last_event.getVoltageValue() * delta_time)
		return total, per_component_consumption


	def calculateCPUEnergy(self,start_time,end_time):
		c_beg_bef,c_beg_aft =  self.__getClosestPair(self.perf_events.events, start_time)
		#c_end_bef,c_end_aft =  getClosestPair(self.perf_events.events, end_time)
		total = 0
		last_event = self.perf_events.events[c_beg_bef]
		last_time = start_time
		tot_time=0
		for i, x in enumerate(self.perf_events.events[c_beg_aft:]):
			if x.time > end_time:
				break
			#print("between %f - %f" %(last_time,x.time) )
			# get different states of cpu btween perf event interval
			#print( x.time - last_time )
			l = self.bat_events.getCPUSamplesInBetween(last_time,x.time)
			# TODO : test to assert if x.time - last_time  = sum( deltas_of_L )
			for sample in l:
				delta, state, voltage = sample[0],sample[1],sample[2]
				cpus_current = last_event.calculateCPUsCurrent(state, self.perf_events.power_profile)
				tot_time += delta
				total += ( cpus_current) * delta * voltage
			last_event = x 
			last_time = x.time
		
		# after calcs'''
		# TODO merge with cycle just like with non cpu
		l = self.bat_events.getCPUSamplesInBetween(last_time, end_time)
		for sample in l:
			delta, state, voltage = sample[0], sample[1], sample[2]
			cpus_current = last_event.calculateCPUsCurrent(state, self.perf_events.power_profile)
			tot_time += delta
			total += (cpus_current) * delta * voltage
		# TODO just like non cpu
		#print(tot_time)
		return total
		

	def __getClosestPair(self,events, time):
		lasti = 0
		for i, x in enumerate(events):
			if x.time>time:
				return lasti, i
			lasti = i
		return lasti,lasti


	def __extractPowerProfile(self, filename):
		# extracting power_profile.xml from device
		res, suc, v = execute_shell_command("adb pull /system/framework/framework-res.apk %s" % self.resources_dir)
		if res==0:
			cmd = """java -jar {res_dir}/apktool_2.4.0.jar d -s {res_dir}/framework-res.apk -f -o {res_dir}/out_jar_dir/""".format(res_dir=self.resources_dir)
			res, suc, v = execute_shell_command(cmd)
			pp_file = self.resources_dir + "/out_jar_dir/res/xml/power_profile.xml"
			if res==0:
				# cp to profiles, remove out_jar_dir and framework-res.apk
				res,_,_=  execute_shell_command("cp {extracted_file} \"{res_dir}/profiles/{new_file}\" ; rm -rf {res_dir}/out_jar_dir {res_dir}/framework-res.apk".format(extracted_file=pp_file, new_file=filename, res_dir=self.resources_dir ))
				if res ==0:
					return filename

		return DEFAULT_PROFILE

	def inferPowerProfile(self):
		res, device_model, _ = execute_shell_command("adb shell getprop ro.product.model")
		if res == 0 and device_model != "":
			model_profile_file = """power_profile_{device_model}.xml""".format(device_model=device_model.replace(" ", "").strip().lower())
			matching_profiles = mega_find(self.resources_dir, pattern=model_profile_file, maxdepth=2)
			if len(matching_profiles) > 0:
				return matching_profiles[0]
			else:
				# if power profile not present in profiles directory, extract from device
				power_profile = self.__extractPowerProfile( model_profile_file)
				print(power_profile)
				return power_profile
		else:
			return DEFAULT_PROFILE

	def __inferTimezone(self):
		res,out,err = execute_shell_command("adb shell date")
		default_tz = DEFAULT_TIMEZONE
		if res == 0 and len(out) > 0:
			default_tz = out.split(" ")[-2]
		log("Using timezone: %s" % default_tz)
		return "UTC" if default_tz == "WEST" else default_tz

	def __unplug_if_fully_charged(self):
		# battery stats file comes empty when battery level == 100
		# using adb to trick device to think it is not charging th battery
		res, o, e = execute_shell_command("adb shell dumpsys battery | grep level | grep 100")
		has_full_charge = res ==0 and "100" in o
		if has_full_charge:
			#mock unplug
			res, o, e = execute_shell_command("adb shell dumpsys battery unplug")
			if res == 0:
				self.unplugged=True
				log("virtually unplugging battery charger while running (battery == 100)", LogSeverity.WARNING)

	def __plug_back(self):
		res, o, e = execute_shell_command("adb shell dumpsys battery reset")
		self.unplugged=False





def have_connected_devices():
	res,o,e = execute_shell_command("adb devices -l | grep -v attached")
	return res == 0 and len(o) > 2


if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument("-p", "--profile", default=None, type=str)
	parser.add_argument("-t", "--timezone", default=None, type=str)
	parser.add_argument("-pft", "--perfettofile", default=None, type=str)
	parser.add_argument("-bts", "--batstatsfile", default=None, type=str)
	args = parser.parse_args()
	has_device_conn = have_connected_devices()
	invalid_file_args = (args.perfettofile is None or args.batstatsfile is None )
	if not has_device_conn and invalid_file_args:
		log("Fatal error. No connected devices and result files submittted for analysis", LogSeverity.FATAL)
		exit(-1)
	g = EManafa(power_profile=args.profile, timezone=args.timezone, resources_dir=MANAFA_RESOURCES_DIR)
	if has_device_conn and invalid_file_args:
		g.init()
		g.start()
		time.sleep(7) # do work
		args.batstatsfile, args.perfettofile = g.stop()
	g.parseResults(args.batstatsfile, args.perfettofile)
	begin = g.bat_events.events[0].time
	end = g.bat_events.events[-1].time
	consumption, per_component_consumption = g.getConsumptionInBetween(begin, end)
	print(per_component_consumption)
	log("Energy consumed: %f Joules" % consumption, log_sev=LogSeverity.SUCCESS)




from abc import ABC, abstractmethod

from manafa.utils.Utils import execute_shell_command, get_results_dir

RESULTS_DIR= get_results_dir()

class Service(ABC):
	"""docstring for Service"""
	def __init__(self, results_dir=""):
		self.results_dir = RESULTS_DIR + "/" + results_dir+"/"

	@abstractmethod
	def config(self, **kwargs):
		print(kwargs)

	@abstractmethod
	def start(self):
		pass

	@abstractmethod
	def stop(self):
		pass

	def clean(self):
		execute_shell_command("find %s -type f | xargs rm " % self.results_dir)



import setuptools
from setuptools import setup

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='manafa',
    version='0.2.99',
    description='E-MANAFA: Energy Monitor and ANAlyzer For Android',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rui Rua',
    author_email='rui.rrua@gmail.com',
    url='https://github.com/RRua/e-manafa',
    license='MIT',
    #package_dir={"": "manafa"},
    packages=setuptools.find_packages(),
    install_requires=[
        'pytz==2021.1',
        'termcolor==1.1.0',
        'python-textops3==3.1.0',
        'python-dateutil==2.8.1',
        "setuptools>=42",
        "wheel"
    ],
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.8',
    ],
    project_urls={
        'Bug Tracker':  'https://github.com/RRua/e-manafa/issues'
        },
    python_requires=">=3.8",
)

'''setup(
    name='e-manafa',
    version='0.1.0',
    packages=[ 'tests.perfetto', 'tests.batterystats', 'manafa', 'manafa.utils', 'manafa.perfetto',
              'manafa.services', 'manafa.batteryStats', 'manafa.powerProfile', 'manafa.resources', 'manafa.results'],
    url='https://github.com/RRua/e-manafa',
    license='MIT',
    author='ruirua',
    author_email='rui.rrua@gmail.com',
    description='e-manafa tool'
)'''



import re,json
#sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
from collections import Iterable

from manafa.powerProfile.PowerProfile import PowerProfile
from manafa.utils.Utils import get_pack_dir
from manafa.utils.dateUtils import convertBatStatTimeToTimeStamp,batStatResetTimeToTimeStamp
import copy

DEFAULT_JSON_PATH = get_pack_dir() +"/batteryStats/BatteryStatus.json"


KNOWN_VALUES=set(
	"screen.on"
	#TODO put here all values in https://source.android.com/devices/tech/power/values
)


def areInMinCPUFreq( cpu_freq , possible_cpu_states  ):
	possible_freqs = possible_cpu_states["speeds"] if "speeds" in possible_cpu_states else []
	if len(possible_freqs)==0:
		possible_freqs = possible_cpu_states["core_speeds"] if "core_speeds" in possible_cpu_states else []
		if len(possible_freqs)>0:
			possible_freqs = sum(possible_freqs.values(),[]) 

	#tem que ser os 2 tempos absolutos
	#print(possible_freqs)
	return False

def safe_division(a,b):
	z = 1 if b==0 else b
	return a / z

class BatteryEvent(object):
	"""docstring for BatteryEvent"""
	def __init__(self,time=0.0,vals={}):
		self.time=time
		self.updates = {}
		self.currents = {}
		self.concurrentUpdates={} 
		self.concurrentUpdates["tmpwhitelist"]=[]
		self.concurrentUpdates["job"]=[]
		self.concurrentUpdates["sync"]=[]
		self.concurrentUpdates["top"]=[]
		self.concurrentUpdates["longwake"]=[]
		self.concurrentUpdates["fg"]=[]
		self.concurrentUpdates["proc"]=[]
		self.concurrentUpdates["screenwake"]=[]
		self.concurrentUpdates["pkgactive"]=[]
		self.concurrentUpdates["user"]=[]
		self.concurrentUpdates["userfg"]=[]
		self.concurrentUpdates["wake_lock_in"]=[]
		self.concurrentUpdates["alarm"]=[]
		self.addEvents(vals)

	def __str__(self):
		return "time:%f vals =  %s , concs= %s  " % (self.time,str(self.updates),str(self.concurrentUpdates))

	def __repr__(self):
		return str(self)

	def isConcurrent(self,state):
		#return re.match(r"(\+|\-)?(tmpwhitelist|job|sync|top|longwake|fg|proc)", state)
		return state in self.concurrentUpdates

	def getCurrentOfBatStatEvent(self):
		currs = {}
		total = 0
		for v, x in self.currents.items():
			try:
				z = float(x)
				currs[v] = z / 1000
				total += z
			except ValueError:
				currs[v]=0
				continue		
		return total/1000, currs

	def getVoltageValue(self):
		return float(self.updates["volt"]) / 1000 if "volt" in self.updates else 0

	def addEvents(self,event1):
		for ev in event1.keys():
			#print("->"+ev + "--")
			if ev.startswith("-"):
				conc_update_state = ev.replace("-","",1)
				if self.isConcurrent(conc_update_state):		
					#print("vou tirar conc "+ev)
					self.concurrentUpdates[conc_update_state] = [n for n in self.concurrentUpdates[conc_update_state] if (  n["val"] != event1[ev]["val"] and n["val2"] != event1[ev]["val2"] ) ]
				else:
					#print("vou tirar ev "+ev)
					self.updates.pop(conc_update_state,None)
			else:
				ev_def = ev.replace("+","",1)
				if self.isConcurrent(ev_def):
					self.concurrentUpdates[ev_def].append(event1[ev])	
				else:
					self.updates[ev_def]=event1[ev]


class BatteryStatsParser(object):
	"""docstring for BatteryStatsParser"""
	def __init__(self, powerProfile=None , timezone="EST", def_file=DEFAULT_JSON_PATH, android_version=10):
		self.events = []
		self.definitions = self.loadDefinitionFile(def_file)
		self.powerProfile = self.loadPowerProfile(powerProfile) if powerProfile is not None else {}
		self.android_version = android_version
		self.start_time = 0
		self.timezone = timezone # adb shell  date +"%Z %z"

	def loadPowerProfile(self, xml_profile ):
		return PowerProfile(xml_profile)

	def loadDefinitionFile(self,def_file):
		with open(def_file,"r") as dff:
			return json.load(dff)

	def getDefinitionVal(self,key,val=""):
		if re.sub(r"\+|\-", "", key) in self.definitions["monoval"]:
			return 0 if "-" in key else 1
		elif re.sub(r"\+|\-", "", key) in self.definitions["trival"]:
			return key
		elif re.sub(r"\+|\-", "", key) in self.definitions["numerical"]:
			return val
		elif re.sub(r"\+|\-", "", key) in self.definitions["nominal"]:
			return self.definitions["nominal"][str(key)][val]
		return None

	def isTrival(self,key):
		return re.sub(r"\+|\-", "", key) in self.definitions["trival"]

	def parseStates(self,states):
		accum=False
		accumulator = ""
		latest_state=None
		events={}
		for state in re.sub(r"^ ","",states).split(" "):
			#print("->" +state)
			if accum:
				accumulator+=state
				if state.count('\"')%2 != 0:
					accum = not accum
					accumulator = ""
				continue		
			if "=" in state:	
				if state.count('\"')%2 != 0:
					accum = True
					accumulator+=state
					continue
				else:
					key = state.split("=")[0]
					val = state.split("=")[1]
					state = self.getDefinitionVal(key,val)
					if self.isTrival(key):
						#print("TODO: split in trival")
						#print("%s - %s -%s" %(key,val.split(":")[0],val.split(":")[1]))
						#return key,val.split(":")[0],val.split(":")[1]
						events[key]= {"val": val.split(":")[0] , "val2": val.split(":")[1]}
					else:
						#print("%s = %s" %(key,state))
						events[key]=state
			elif state != "" and state != " ":
				st = self.getDefinitionVal(state)
				#print("%s - %s" %(state,st))
				#print("val "+str(self.getDefinitionVal(state)))
				events[state]=st		
		return events
	
	def parseHistory(self,lines_list):
		for i, line in enumerate(lines_list):
			if re.match(r"^Battery History \([0-9]", line):
				#header, ignore
				continue
			if re.match(r"^Per-PID Stats", line)or len(line)==0:
				return
			elif re.match(r"^\s*([^\s]+) (\(\d+\)) (\d+)(.*)?$", line):
				x = re.match(r"^\s*([^\s]+) (\(\d+\)) (\d+)(.*)?$", line)
				time   = convertBatStatTimeToTimeStamp(x.groups()[0], timezone=self.timezone)
				time  += self.start_time
				#print(time)
				events = self.parseStates(x.groups()[3])
				self.addUpdate(time, events)
			elif re.match(r"^\s*0 (\(\d+\)) (.*)?$", line):
				x = re.match(r"^\s*0 (\(\d+\)) (.*)?$", line)
				if "RESET:TIME" in x.groups()[1]:
					self.start_time= batStatResetTimeToTimeStamp((x.groups()[1]).replace("RESET:TIME: ",""), self.timezone )		
					#print(epochToDate(self.start_time))
			else:
				# TODO Handle DcpuStats and DpstStats
				print("linha invalida" + line)

	def addUpdate(self, time,bat_events):
		if len(self.events)==0:
			bt = BatteryEvent( time,bat_events )
			self.estimateCurrentConsumption(bt)
			self.events.append(bt )
		else:	
			last_added = self.events[-1]
			if last_added.time == time:
				self.events[-1].addEvents(bat_events)
				
			else:
				# todo try to replace with shallow copy
				bt = copy.deepcopy(self.events[-1])
				bt.time = time
				bt.addEvents(bat_events)
				self.estimateCurrentConsumption(bt)
				self.events.append(bt)


	def estimateCurrentConsumption(self,bt_event):
		power={}
		for p,v in self.powerProfile.components.items():
			st = self.determinateComponentCurrent(bt_event,p,v)
			power[p]=st
			#print("%s %s" %(p , str(st)))
		bt_event.currents=power
		

	def getCPUSamplesInBetween(self,start_time, end_time ):
		l =[]
		last_ev = self.events[0] if len(self.events)>0 else None
		last_time = start_time
		for x in self.events:
			if x.time > start_time and x.time < end_time:
				delta = x.time - last_time
				state = last_ev.currents["cpu"]
				voltage = last_ev.getVoltageValue() #float(last_ev.updates["volt"])
				pair = (delta, state,voltage)
				l.append(pair)
				last_time= x.time
			last_ev = x
		# 
		last_delta = end_time - last_time
		last_state = last_ev.currents["cpu"]
		last_voltage = last_ev.getVoltageValue() #float(last_ev.updates["volt"])
		last_pair = ( last_delta, last_state,last_voltage)
		l.append(last_pair)
		return l

	def determinateComponentCurrent(self,bt_event,comp_name, possible_states):
		current = 0.0
		curravg=0
		avg_ct=0
		# screen
		if comp_name == "screen" and "screen" in bt_event.updates:
				on_current = possible_states["on"]
				brightness_level = bt_event.updates["brightness"] if "brightness" in bt_event.updates else 1
				relative_full_current = ( brightness_level * possible_states["full"] / ( len( self.definitions["nominal"]["brightness"] ) -1) ) 
				current+= on_current + relative_full_current 

		elif comp_name == "ambient" and "screen_doze" in bt_event.updates:
				# power profile might have a defined value for ambient/doze screen consumpri
				doze_current = possible_states["on"]
				current+=doze_current

		# camera/flashlight 
		elif comp_name == "camera":
			if "camera" in bt_event.updates:
				#usually available as avg consumption (Intended as a rough estimate for an application running a preview and capturing approximately 10 full-resolution pictures per minute.)
				cam_current = possible_states["avg"]
				current+=cam_current
		
			if "flashlight" in bt_event.updates:
				flash_curr = possible_states["flashlight"]
				current+=flash_curr
		# dsp
		elif comp_name == "dsp":
			if "video" in bt_event.updates:
				video_curr = possible_states["video"] if comp_name == "dsp" else possible_states
				current+=video_curr

			if "audio" in bt_event.updates:
				
				audio_curr = possible_states["audio"] if comp_name == "dsp" else possible_states
				current+=audio_curr
		
		# audio
		elif comp_name == "video" and "video" in bt_event.updates:
			video_curr = possible_states
			current += video_curr
		# video
		elif comp_name == "audio" and "audio" in bt_event.updates:
			audio_curr = possible_states
			current += audio_curr

		# wifi 
		elif comp_name == "wifi" and "wifi_running" in bt_event.updates:
			on_current = possible_states["on"] if "on" in possible_states else 0
			on_current += possible_states["controller"]["idle"] if ( "controller" in possible_states and "idle" in possible_states["controller"] ) else 0
			current+= on_current
			if "wifi_scan" in bt_event.updates:
				if "scan" in possible_states:
					current+= possible_states["scan"]
				if "controller" in possible_states:
					curravg=0
					avg_ct =0
					if "tx" in possible_states["controller"]:
						curravg+=possible_states["controller"]["tx"]
						avg_ct+=1
					if "rx" in possible_states["controller"]:
						curravg+=possible_states["controller"]["rx"]
						avg_ct+=1
					current += safe_division(curravg, avg_ct)
			elif "wifi_radio" in bt_event.updates:
				current+= possible_states["active"] if "active" in possible_states else 0
				if "controller" in possible_states:
					curravg=0
					avg_ct =0
					if "tx" in possible_states["controller"]:
						curravg+=possible_states["controller"]["tx"]
						avg_ct+=1
					if "rx" in possible_states["controller"]:
						curravg+=possible_states["controller"]["rx"]
						avg_ct+=1
					current += safe_division(curravg, avg_ct)

		# gps 
		elif comp_name == "gps":
			if "signalqualitybased" in possible_states: # and "gps_signal_quality" in bt_event.updates:
				# considerate gps signal quality
				if "gps_signal_quality" in bt_event.updates:
					val = 1 if bt_event.updates["gps_signal_quality"] == "good" else 0
					current+= possible_states["signalqualitybased"][val]
			if "on" in 	possible_states and "gps" in bt_event.updates:
				current+= possible_states["on"]
		
		#bluetooth
		elif comp_name == "bluetooth":
			if self.android_version<7 or "controller" not in possible_states:
				#account blue on and active vals
				if "ble_scan" in  bt_event.updates:
					current+=  possible_states["active"] if "active" in possible_states else 0
					current+=  possible_states["on"] if "on" in possible_states else 0
				elif "bluetooth" in  bt_event.updates:
					current+=  possible_states["on"] if "on" in possible_states else 0
			elif "controller" in possible_states:
				current += possible_states["controller"]["idle"] if ( "idle" in possible_states["controller"] ) else 0
				if "ble_scan" in  bt_event.updates:
					if "tx" in possible_states["controller"]:
						curravg+=possible_states["controller"]["tx"]
						avg_ct+=1
					if "rx" in possible_states["controller"]:
						curravg+=possible_states["controller"]["rx"]
						avg_ct+=1
					current += safe_division(curravg, avg_ct)

		# radio =  modem
		elif comp_name == "radio":
			#radio.on
			if "phone_scanning" in  bt_event.updates:
				# radio.scanning
				current += possible_states["scanning"] if "scanning" in possible_states else 0

			elif "mobile_radio" in  bt_event.updates:
				on_vals =  list(possible_states["on"]) if "on" in possible_states else []
				signal_stren = bt_event.updates["phone_signal_strength"] if "phone_signal_strength" in bt_event.updates else 0
				if signal_stren > len(on_vals) and len(on_vals)>0:
					 current+= on_vals[-1]
				elif len(on_vals)>0:
					current+= on_vals[signal_stren]
				# radio.active == mobile_radio - transmiting
				
		elif comp_name == "modem":
			#same as radio
			if "phone_scanning" in  bt_event.updates:
				curravg=0
				avg_ct=0
				if "tx" in possible_states["controller"]:
					on_vals =  possible_states["controller"]["tx"] if "tx" in possible_states["controller"] else [] 
					signal_stren = bt_event.updates["phone_signal_strength"] if "phone_signal_strength" in bt_event.updates else 0
					if signal_stren > len(on_vals) and len(on_vals)>0:
						curravg+= on_vals[-1]
					elif  len(on_vals)>0:
						curravg+= on_vals[signal_stren]

					avg_ct+=1
				if "rx" in possible_states["controller"]:
					curravg+=possible_states["controller"]["rx"]
					avg_ct+=1
				current += safe_division(curravg, avg_ct)
			elif "mobile_radio" in  bt_event.updates:
				current+=  possible_states["idle"] if "idle" in possible_states else 0	
		# cpu
		elif comp_name == "cpu":
			# retrieve just the component state
			#if phone has multiple cpu_clusters and that info is present in power profile file
			# only for devices with heterogeneous CPU architectures.
			#print(possible_states)
			#has_multiple_cpu_clusters= "clusters" in possible_states and "cores" in possible_states["clusters"] and  len(possible_states["clusters"]["cores"])>1
			#if has_multiple_cpu_clusters:
				#cores_per_cluster = possible_states["clusters"]["cores"]
				# calculate energy per cluster
				#if areInMinCPUFreq(bt_event.updates["cpufreq"] , possible_states):
						# assume is just awake

			#		current += possible_states["awake"]
		#		else:

					# calculate active state
		#			current += 0 # possible_states["awake"] if "awake" in possible_states else 0
		#			print("TODO Calculate power according to freq")
				


		#	else:
				if "running" in bt_event.updates:
					# is active or just awake	
					#if "awake" in possible_states and areInMinCPUFreq(bt_event.updates["cpufreq"] , possible_states):
					#if areInMinCPUFreq(bt_event.updates["cpufreq"] , possible_states):
						# assume is just awake

					#	current += possible_states["awake"]
					#else:
						# calculate active state
					#	current += 0 # possible_states["awake"] if "awake" in possible_states else 0
					#	print("TODO Calculate power according to freq")
					current = "active"
				else:
					#cpu in idle state
					#current+= possible_states["idle"] if "idle" in possible_states else 0
					current = "idle"

		return current

	def parseFile(self,filename):
		with open(filename, 'r') as filehandle:
			lines=filehandle.read().splitlines()
			self.parseHistory(lines)

#if __name__ == '__main__':
	#if len(sys.argv)>1:
		#pp = "samples/profiles/power_profile.xml"
		#pp = "samples/profiles/power_profile_pixel3a_grapheneos.xml"
		#x = BatteryStatsParser(powerProfile=pp,timezone="WET")
		#x.parseFile(sys.argv[1])
		






# Battery related : https://android.googlesource.com/platform/frameworks/base/+/master/core/java/android/os/BatteryManager.java

# https://android.googlesource.com/platform/hardware/interfaces/+/oreo-m5-release/health/1.0/types.hal
BATTERY_STATUS= {
 "unknown" :   1, # "BATTERY_STATUS_UNKNOWN",
 "charging":    2, #  "BATTERY_STATUS_CHARGING",
 "discharging":   3, #; "BATTERY_STATUS_DISCHARGING",
 "not-charging":  4, # "BATTERY_STATUS_NOT_CHARGING",
 "full":  5 # "BATTERY_STATUS_FULL",
}
BATTERY_HEALTH= {
    "unknown":1, # "BATTERY_HEALTH_UNKNOWN",
    "good":2, # "BATTERY_HEALTH_GOOD",
    "overheat":3, # "BATTERY_HEALTH_OVERHEAT",
    "dead":4, # "BATTERY_HEALTH_DEAD",
    "over-voltage":5, # "BATTERY_HEALTH_OVER_VOLTAGE",
    "failure":6, # "BATTERY_HEALTH_UNSPECIFIED_FAILURE",
    "cold":7, # "BATTERY_HEALTH_COLD",
}

BATTERY_PLUGGED= {
    "none": 0 , # : "BATTERY_PLUGGED_NONE"
    "ac": 1 , #  "BATTERY_PLUGGED_AC",
    "usb": 2 , #  "BATTERY_PLUGGED_USB",
    "wireless": 4 , #  "BATTERY_PLUGGED_WIRELESS",
}



# Battery related : https://android.googlesource.com/platform/frameworks/base/+/master/core/java/android/os/BatteryManager.java

# https://android.googlesource.com/platform/hardware/interfaces/+/oreo-m5-release/health/1.0/types.hal


#Display
#https://android.googlesource.com/platform/frameworks/base/+/master/core/java/android/view/Display.java
DISPLAY_STATE= {
    1: "STATE_OFF",
    2: "STATE_ON",
    3: "STATE_DOZE",
    4: "STATE_DOZE_SUSPEND",
    5: "STATE_DOZE_VR",
    6: "DISPLAY_STATE_ON_SUSPEND"
}

GENERIC_CONSTANTS = {   
        1  :"WAKE_TYPE_FULL" ,
        2  :"WAKE_TYPE_WINDOW" ,
        3  :"SENSOR" ,
        4  :"WIFI_RUNNING" ,
        5  :"FULL_WIFI_LOCK" ,
        6  :"WIFI_SCAN" ,
        7  :"WIFI_MULTICAST_ENABLED" ,
        8  :"VIDEO_TURNED_ON" ,
        9  :"VIBRATOR_ON" ,
        10 :"FOREGROUND_ACTIVITY" ,
        11 :"WIFI_BATCHED_SCAN" ,
        12 :"PROCESS_STATE" ,
        13 :"SYNC" ,
        14 :"JOB" ,
        15 :"AUDIO_TURNED_ON" ,
        16 :"FLASHLIGHT_TURNED_ON" ,
        17 :"CAMERA_TURNED_ON" ,
        18 :"WAKE_TYPE_DRAW" ,
        19 :"BLUETOOTH_SCAN_ON" ,
        20 :"AGGREGATED_WAKE_TYPE_PARTIAL" ,
        21 :"BLUETOOTH_UNOPTIMIZED_SCAN_ON" ,
        22 :"FOREGROUND_SERVICE" ,
        23 :"WIFI_AGGREGATE_MULTICAST_ENABLED" ,
}

STATS={
    0 : "STATS_SINCE_CHARGED",
    1 : "STATS_CURRENT",
    2 : "STATS_SINCE_UNPLUGGED"
}

EVENT_NAMES= {
            "null", 
            "proc", 
            "fg", 
            "top", 
            "sync", 
            "wake_lock_in", 
            "job", 
            "user", 
            "userfg", 
            "conn",
            "active", 
            "pkginst", 
            "pkgunin", 
            "alarm", 
            "stats", 
            "pkginactive", 
            "pkgactive",
            "tmpwhitelist", 
            "screenwake", 
            "wakeupap", 
            "longwake", 
            "est_capacity"
}



import os,re
import datetime, pytz,time



def convertToUnixTimestamp(batstattime):
	return batstattime	

def convertBatStatTimeToTimeStamp(batstattime,timezone="EST"):
	#secs=[86400,3600,60,1,0.001]
	secs=[0.001,1,60,3600,86400]
	val=0.0
	ts = re.split(r'[a-z]+',batstattime.strip())
	ts.reverse()
	ts = filter (lambda x:x!='', ts)
	for i,x in enumerate(ts):
		#print(x)
		val += float(x) * secs[i]
		#print(val)
	return val
	#d1 =datetime.datetime.strptime(batstattime, '%Y-%m-%d-%H-%M-%S')
	#pst = pytz.timezone(timezone)
	#d= pst.localize(d1)
	#return time.mktime(d.timetuple())

def batStatResetTimeToTimeStamp(matime, timezone="UTC"):
	d1 =datetime.datetime.strptime(matime, '%Y-%m-%d-%H-%M-%S')
	pst = pytz.timezone(timezone)
	d= pst.localize(d1)
	return time.mktime(d.timetuple())


def convertDateToTimeStamp(date,timezone="EST"):
	local = pytz.timezone(timezone)
	#time = time.mktime(datetime.datetime.strptime(s, "%d/%m/%Y").timetuple())
	return time
	#return None

def epochToDate(ts):
	return time.ctime(ts)

#x="2020-11-17-12-06-18"
#z = batStatResetTimeToTimeStamp(x)
#print(z)
#t2 = "1s394ms"
#t2 = "3d23h23m38s126ms"
#zz = convertBatStatTimeToTimeStamp(t2)
#print(zz)
#z = z+zz
#print(z)


